@@ -19,12 +19,13 @@
 #include "spice-common.h"
 #include "spice-session-priv.h"
 #include "spice-channel-priv.h"
+#include "spice-util-priv.h"
 
 #include <pulse/glib-mainloop.h>
 #include <pulse/pulseaudio.h>
 
 #define SPICE_PULSE_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_PULSE, spice_pulse))
+    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_PULSE, SpicePulsePrivate))
 
 struct stream {
     pa_sample_spec          spec;
@@ -35,8 +36,7 @@ struct stream {
     gboolean                started;
 };
 
-struct spice_pulse {
-    SpiceSession            *session;
+struct _SpicePulsePrivate {
     SpiceChannel            *pchannel;
     SpiceChannel            *rchannel;
 
@@ -72,12 +72,11 @@ static const char *context_state_names[] = {
 static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
                           gpointer data);
 static void stream_stop(SpicePulse *pulse, struct stream *s);
-static void channel_new(SpiceSession *s, SpiceChannel *channel,
-                        gpointer data);
+static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel);
 
 static void spice_pulse_finalize(GObject *obj)
 {
-    spice_pulse *p;
+    SpicePulsePrivate *p;
 
     p = SPICE_PULSE_GET_PRIVATE(obj);
 
@@ -92,7 +91,7 @@ static void spice_pulse_finalize(GObject *obj)
 
 static void spice_pulse_dispose(GObject *obj)
 {
-    spice_pulse *p;
+    SpicePulsePrivate *p;
 
     SPICE_DEBUG("%s", __FUNCTION__);
     p = SPICE_PULSE_GET_PRIVATE(obj);
@@ -113,46 +112,33 @@ static void spice_pulse_dispose(GObject *obj)
         pa_operation_unref(p->record.cork_op);
     p->record.cork_op = NULL;
 
-    if (p->pchannel != NULL) {
-        g_signal_handlers_disconnect_by_func(p->pchannel,
-                                             channel_event, obj);
+    if (p->pchannel)
         g_object_unref(p->pchannel);
-        p->pchannel = NULL;
-    }
+    p->pchannel = NULL;
 
-    if (p->rchannel != NULL) {
-        g_signal_handlers_disconnect_by_func(p->rchannel,
-                                             channel_event, obj);
+    if (p->rchannel)
         g_object_unref(p->rchannel);
-        p->rchannel = NULL;
-    }
-
-    if (p->session != NULL) {
-        g_signal_handlers_disconnect_by_func(p->session,
-                                             channel_new, obj);
-        g_object_unref(p->session);
-        p->session = NULL;
-    }
+    p->rchannel = NULL;
 
     G_OBJECT_CLASS(spice_pulse_parent_class)->dispose(obj);
 }
 
 static void spice_pulse_init(SpicePulse *pulse)
 {
-    spice_pulse *p;
-
-    p = pulse->priv = SPICE_PULSE_GET_PRIVATE(pulse);
-    memset(p, 0, sizeof(*p));
+    pulse->priv = SPICE_PULSE_GET_PRIVATE(pulse);
 }
 
 static void spice_pulse_class_init(SpicePulseClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
+    SpiceAudioClass *audio_class = SPICE_AUDIO_CLASS(klass);
+
+    audio_class->connect_channel = connect_channel;
 
     gobject_class->finalize = spice_pulse_finalize;
     gobject_class->dispose = spice_pulse_dispose;
 
-    g_type_class_add_private(klass, sizeof(spice_pulse));
+    g_type_class_add_private(klass, sizeof(SpicePulsePrivate));
 }
 
 /* ------------------------------------------------------------------ */
@@ -169,7 +155,7 @@ static void pulse_uncork_cb(pa_stream *pastream, int success, void *data)
 
 static void stream_uncork(SpicePulse *pulse, struct stream *s)
 {
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_operation *o = NULL;
 
     if (s->cork_op) {
@@ -214,7 +200,7 @@ static void pulse_cork_cb(pa_stream *pastream, int success, void *data)
 
 static void stream_cork(SpicePulse *pulse, struct stream *s)
 {
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_operation *o = NULL;
 
     if (s->uncork_op) {
@@ -234,7 +220,7 @@ static void stream_cork(SpicePulse *pulse, struct stream *s)
 
 static void stream_stop(SpicePulse *pulse, struct stream *s)
 {
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
 
     if (pa_stream_disconnect(s->stream) < 0) {
         g_warning("pa_stream_disconnect() failed: %s",
@@ -246,7 +232,7 @@ static void stream_stop(SpicePulse *pulse, struct stream *s)
 
 static void stream_state_callback(pa_stream *s, void *userdata)
 {
-    spice_pulse *p;
+    SpicePulsePrivate *p;
     p = SPICE_PULSE_GET_PRIVATE(userdata);
 
     g_return_if_fail(p != NULL);
@@ -285,7 +271,7 @@ static void stream_underflow_cb(pa_stream *s, void *userdata)
 
 static void create_playback(SpicePulse *pulse)
 {
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_stream_flags_t flags;
     pa_buffer_attr buffer_attr = { 0, };
 
@@ -318,7 +304,7 @@ static void playback_start(SpicePlaybackChannel *channel, gint format, gint chan
                            gint frequency, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_context_state_t state;
 
     g_return_if_fail(p != NULL);
@@ -362,7 +348,7 @@ static void playback_data(SpicePlaybackChannel *channel,
                           gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_stream_state_t state;
 
     if (!p->playback.stream)
@@ -395,7 +381,7 @@ static void playback_data(SpicePlaybackChannel *channel,
 static void playback_stop(SpicePlaybackChannel *channel, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
 
     SPICE_DEBUG("%s", __FUNCTION__);
 
@@ -410,7 +396,7 @@ static void stream_update_timing_callback(pa_stream *s, int success, void *userd
 {
     pa_usec_t usec;
     int negative = 0;
-    spice_pulse *p;
+    SpicePulsePrivate *p;
 
     p = SPICE_PULSE_GET_PRIVATE(userdata);
 
@@ -434,7 +420,7 @@ static void stream_update_timing_callback(pa_stream *s, int success, void *userd
 static void playback_get_delay(SpicePlaybackChannel *channel, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
 
     if (p->playback.stream && pa_stream_get_state(p->playback.stream) == PA_STREAM_READY) {
         pa_operation *o;
@@ -449,7 +435,7 @@ static void playback_get_delay(SpicePlaybackChannel *channel, gpointer data)
 static void stream_read_callback(pa_stream *s, size_t length, void *data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
 
     g_return_if_fail(p != NULL);
 
@@ -481,7 +467,7 @@ static void stream_read_callback(pa_stream *s, size_t length, void *data)
 
 static void create_record(SpicePulse *pulse)
 {
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_buffer_attr buffer_attr = { 0, };
     pa_stream_flags_t flags;
 
@@ -513,7 +499,7 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
                          gint frequency, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
     pa_context_state_t state;
 
     p->record.started = TRUE;
@@ -553,7 +539,7 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
 static void record_stop(SpiceRecordChannel *channel, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
 
     SPICE_DEBUG("%s", __FUNCTION__);
 
@@ -568,7 +554,7 @@ static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
                           gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
 
     switch (event) {
     case SPICE_CHANNEL_OPENED:
@@ -594,7 +580,7 @@ static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
 static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
     guint16 *volume;
     guint nchannels;
     pa_operation *op;
@@ -630,7 +616,7 @@ static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer
 static void playback_mute_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
     gboolean mute;
     pa_operation *op;
 
@@ -654,7 +640,7 @@ static void playback_mute_changed(GObject *object, GParamSpec *pspec, gpointer d
 static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
     gboolean mute;
     pa_operation *op;
 
@@ -678,7 +664,7 @@ static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer dat
 static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
     SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulsePrivate *p = pulse->priv;
     guint16 *volume;
     guint nchannels;
     pa_operation *op;
@@ -712,51 +698,57 @@ static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer d
         pa_operation_unref(op);
 }
 
-static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
+static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
 {
-    SpicePulse *pulse = data;
-    spice_pulse *p = pulse->priv;
+    SpicePulse *pulse = SPICE_PULSE(audio);
+    SpicePulsePrivate *p = pulse->priv;
 
     if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
-        g_return_if_fail(p->pchannel == NULL);
+        g_return_val_if_fail(p->pchannel == NULL, FALSE);
+
         p->pchannel = g_object_ref(channel);
-        g_signal_connect(channel, "playback-start",
-                         G_CALLBACK(playback_start), pulse);
-        g_signal_connect(channel, "playback-get-delay",
-                         G_CALLBACK(playback_get_delay), pulse);
-        g_signal_connect(channel, "playback-data",
-                         G_CALLBACK(playback_data), pulse);
-        g_signal_connect(channel, "playback-stop",
-                         G_CALLBACK(playback_stop), pulse);
-        g_signal_connect(channel, "channel-event",
-                         G_CALLBACK(channel_event), pulse);
-        g_signal_connect(channel, "notify::volume",
-                         G_CALLBACK(playback_volume_changed), pulse);
-        g_signal_connect(channel, "notify::mute",
-                         G_CALLBACK(playback_mute_changed), pulse);
-        spice_channel_connect(channel);
+        spice_g_signal_connect_object(channel, "playback-start",
+                                      G_CALLBACK(playback_start), pulse, 0);
+        spice_g_signal_connect_object(channel, "playback-get-delay",
+                                      G_CALLBACK(playback_get_delay), pulse, 0);
+        spice_g_signal_connect_object(channel, "playback-data",
+                                      G_CALLBACK(playback_data), pulse, 0);
+        spice_g_signal_connect_object(channel, "playback-stop",
+                                      G_CALLBACK(playback_stop), pulse, 0);
+        spice_g_signal_connect_object(channel, "channel-event",
+                                      G_CALLBACK(channel_event), pulse, 0);
+        spice_g_signal_connect_object(channel, "notify::volume",
+                                      G_CALLBACK(playback_volume_changed), pulse, 0);
+        spice_g_signal_connect_object(channel, "notify::mute",
+                                      G_CALLBACK(playback_mute_changed), pulse, 0);
+
+        return TRUE;
     }
 
     if (SPICE_IS_RECORD_CHANNEL(channel)) {
-        g_return_if_fail(p->rchannel == NULL);
+        g_return_val_if_fail(p->rchannel == NULL, FALSE);
+
         p->rchannel = g_object_ref(channel);
-        g_signal_connect(channel, "record-start",
-                         G_CALLBACK(record_start), pulse);
-        g_signal_connect(channel, "record-stop",
-                         G_CALLBACK(record_stop), pulse);
-        g_signal_connect(channel, "channel-event",
-                         G_CALLBACK(channel_event), pulse);
-        g_signal_connect(channel, "notify::volume",
-                         G_CALLBACK(record_volume_changed), pulse);
-        g_signal_connect(channel, "notify::mute",
-                         G_CALLBACK(record_mute_changed), pulse);
-        spice_channel_connect(channel);
+        spice_g_signal_connect_object(channel, "record-start",
+                                      G_CALLBACK(record_start), pulse, 0);
+        spice_g_signal_connect_object(channel, "record-stop",
+                                      G_CALLBACK(record_stop), pulse, 0);
+        spice_g_signal_connect_object(channel, "channel-event",
+                                      G_CALLBACK(channel_event), pulse, 0);
+        spice_g_signal_connect_object(channel, "notify::volume",
+                                      G_CALLBACK(record_volume_changed), pulse, 0);
+        spice_g_signal_connect_object(channel, "notify::mute",
+                                      G_CALLBACK(record_mute_changed), pulse, 0);
+
+        return TRUE;
     }
+
+    return FALSE;
 }
 
 static void context_state_callback(pa_context *c, void *userdata)
 {
-    spice_pulse *p;
+    SpicePulsePrivate *p;
     p = SPICE_PULSE_GET_PRIVATE(userdata);
 
     g_return_if_fail(p != NULL);
@@ -793,20 +785,13 @@ SpicePulse *spice_pulse_new(SpiceSession *session, GMainContext *context,
                             const char *name)
 {
     SpicePulse *pulse;
-    spice_pulse *p;
-    GList *list, *tmp;
+    SpicePulsePrivate *p;
 
-    pulse = g_object_new(SPICE_TYPE_PULSE, NULL);
+    pulse = g_object_new(SPICE_TYPE_PULSE,
+                         "session", session,
+                         "main-context", context,
+                         NULL);
     p = SPICE_PULSE_GET_PRIVATE(pulse);
-    p->session = g_object_ref(session);
-
-    g_signal_connect(session, "channel-new",
-                     G_CALLBACK(channel_new), pulse);
-    list = spice_session_get_channels(session);
-    for (tmp = g_list_first(list); tmp != NULL; tmp = g_list_next(tmp)) {
-        channel_new(session, tmp->data, (gpointer)pulse);
-    }
-    g_list_free(list);
 
     p->mainloop = pa_glib_mainloop_new(context);
     p->state = PA_CONTEXT_READY;
