@@ -22,6 +22,7 @@
 #include <gio/gio.h>
 #include "desktop-integration.h"
 #include "spice-session.h"
+#include "spice-proxy.h"
 #include "spice-gtk-session.h"
 #include "spice-channel-cache.h"
 #include "decode.h"
@@ -44,6 +45,7 @@ struct _SpiceSessionPrivate {
     char              *cert_subject;
     guint             verify;
     gboolean          read_only;
+    SpiceProxy        *proxy;
 
     /* whether to enable audio */
     gboolean          audio;
