@@ -106,6 +106,7 @@ enum {
     PROP_UUID,
     PROP_NAME,
     PROP_CA,
+    PROP_PROXY
 };
 
 /* signals */
@@ -118,6 +119,33 @@ enum {
 static guint signals[SPICE_SESSION_LAST_SIGNAL];
 
 
+static void update_proxy(SpiceSession *self, const gchar *str)
+{
+    SpiceSessionPrivate *s = self->priv;
+    SpiceProxy *proxy = NULL;
+    GError *error = NULL;
+
+    if (str == NULL)
+        str = g_getenv("SPICE_PROXY");
+    if (str == NULL || *str == 0) {
+        g_clear_object(&s->proxy);
+        return;
+    }
+
+    proxy = spice_proxy_new();
+    if (!spice_proxy_parse(proxy, str, &error))
+        g_clear_object(&proxy);
+    if (error) {
+        g_warning("%s", error->message);
+        g_clear_error(&error);
+    }
+
+    if (proxy != NULL) {
+        g_clear_object(&s->proxy);
+        s->proxy = proxy;
+    }
+}
+
 static void spice_session_init(SpiceSession *session)
 {
     SpiceSessionPrivate *s;
@@ -147,6 +175,7 @@ static void spice_session_init(SpiceSession *session)
     cache_init(&s->images, "image");
     cache_init(&s->palettes, "palette");
     s->glz_window = glz_decoder_window_new();
+    update_proxy(session, NULL);
 }
 
 static void
@@ -179,6 +208,7 @@ spice_session_dispose(GObject *gobject)
     g_clear_object(&s->desktop_integration);
     g_clear_object(&s->gtk_session);
     g_clear_object(&s->usb_manager);
+    g_clear_object(&s->proxy);
 
     /* Chain up to the parent class */
     if (G_OBJECT_CLASS(spice_session_parent_class)->dispose)
@@ -491,6 +521,9 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_UUID:
         g_value_set_pointer(value, s->uuid);
 	break;
+    case PROP_PROXY:
+        g_value_take_string(value, spice_proxy_to_string(s->proxy));
+	break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
 	break;
@@ -606,6 +639,9 @@ static void spice_session_set_property(GObject      *gobject,
         g_clear_pointer(&s->ca, g_byte_array_unref);
         s->ca = g_value_dup_boxed(value);
         break;
+    case PROP_PROXY:
+        update_proxy(session, g_value_get_string(value));
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -894,7 +930,7 @@ static void spice_session_class_init(SpiceSessionClass *klass)
      * hardware smartcard reader. If it's set to a NULL-terminated string
      * array containing the names of 3 valid certificates, these will be
      * used to simulate a smartcard in the guest
-     * @see_also: spice_smartcard_manager_insert_card()
+     * See also spice_smartcard_manager_insert_card()
      *
      * Since: 0.7
      **/
@@ -944,7 +980,7 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                           G_PARAM_STATIC_STRINGS));
 
     /**
-     * SpiceSession::inhibit-keyboard-grab
+     * SpiceSession::inhibit-keyboard-grab:
      *
      * This boolean is set by the usbredir channel to indicate to #SpiceDisplay
      * that the keyboard grab should be temporarily released, because it is
@@ -1097,6 +1133,23 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                               G_PARAM_READABLE |
                               G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:proxy:
+     *
+     * URI to the proxy server to use when doing network connection.
+     * of the form <![CDATA[ [protocol://]<host>[:port] ]]>
+     *
+     * Since: 0.17
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_PROXY,
+         g_param_spec_string("proxy",
+                             "Proxy",
+                             "The proxy server",
+                             NULL,
+                             G_PARAM_READWRITE |
+                             G_PARAM_STATIC_STRINGS));
+
     g_type_class_add_private(klass, sizeof(SpiceSessionPrivate));
 }
 
@@ -1647,21 +1700,6 @@ static void proxy_lookup_ready(GObject *source_object, GAsyncResult *result,
     open_host_connectable_connect(open_host, G_SOCKET_CONNECTABLE(address));
     g_resolver_free_addresses(addresses);
 }
-
-static SpiceProxy* get_proxy(GError **error)
-{
-    SpiceProxy *proxy;
-
-    const gchar *proxy_env = g_getenv("SPICE_PROXY");
-    if (proxy_env == NULL || strlen(proxy_env) == 0)
-        return NULL;
-
-    proxy = spice_proxy_new();
-    if (!spice_proxy_parse(proxy, proxy_env, error))
-        g_clear_object(&proxy);
-
-    return proxy;
-}
 #endif
 
 /* main context */
@@ -1674,24 +1712,28 @@ static gboolean open_host_idle_cb(gpointer data)
     g_return_val_if_fail(open_host->socket == NULL, FALSE);
 
 #if GLIB_CHECK_VERSION(2,26,0)
-    open_host->proxy = get_proxy(&open_host->error);
-    if (open_host->proxy) {
+    open_host->proxy = s->proxy;
+    if (open_host->error != NULL) {
+        coroutine_yieldto(open_host->from, NULL);
+        return FALSE;
+    }
+
+    if (open_host->proxy)
         g_resolver_lookup_by_name_async(g_resolver_get_default(),
                                         spice_proxy_get_hostname(open_host->proxy),
                                         open_host->cancellable,
                                         proxy_lookup_ready, open_host);
-    } else
+    else
 #endif
-    if (open_host->error != NULL) {
-        coroutine_yieldto(open_host->from, NULL);
-        return FALSE;
-    } else
         open_host_connectable_connect(open_host,
                                       g_network_address_new(s->host, open_host->port));
 
     SPICE_DEBUG("open host %s:%d", s->host, open_host->port);
-    if (open_host->proxy != NULL)
-        SPICE_DEBUG("(with proxy %p)", open_host->proxy);
+    if (open_host->proxy != NULL) {
+        gchar *str = spice_proxy_to_string(open_host->proxy);
+        SPICE_DEBUG("(with proxy %s)", str);
+        g_free(str);
+    }
 
     return FALSE;
 }
@@ -1728,7 +1770,6 @@ GSocket* spice_session_channel_open_host(SpiceSession *session, SpiceChannel *ch
     }
 
     g_clear_object(&open_host.client);
-    g_clear_object(&open_host.proxy);
     return open_host.socket;
 }
 
