@@ -22,6 +22,9 @@
 #include <glib/gi18n.h>
 
 #include <sys/stat.h>
+#ifdef HAVE_TERMIOS_H
+#include <termios.h>
+#endif
 
 #define GNOME_DESKTOP_USE_UNSTABLE_API 2
 #include "display/gnome-rr.h"
@@ -128,6 +131,7 @@ static char *spicy_title = NULL;
 static GMainLoop     *mainloop = NULL;
 static int           connections = 0;
 static GKeyFile      *keyfile = NULL;
+static SpicePortChannel*stdin_port = NULL;
 static GnomeRRScreen *rrscreen = NULL;
 static GnomeRRConfig *rrsaved = NULL;
 static GnomeRRConfig *rrcurrent = NULL;
@@ -305,17 +309,23 @@ static int connect_dialog(SpiceSession *session)
 
 static void update_status_window(SpiceWindow *win)
 {
-    char status[256];
+    gchar *status;
 
     if (win == NULL)
         return;
+
     if (win->mouse_grabbed) {
-        snprintf(status, sizeof(status), _("Use Shift+F12 to ungrab mouse."));
+        SpiceGrabSequence *sequence = spice_display_get_grab_keys(SPICE_DISPLAY(win->spice));
+        gchar *seq = spice_grab_sequence_as_string(sequence);
+        status = g_strdup_printf(_("Use %s to ungrab mouse."), seq);
+        g_free(seq);
     } else {
-        snprintf(status, sizeof(status), _("mouse: %s, agent: %s"),
+        status = g_strdup_printf(_("mouse: %s, agent: %s"),
                  win->conn->mouse_state, win->conn->agent_state);
     }
+
     gtk_label_set_text(GTK_LABEL(win->status), status);
+    g_free(status);
 }
 
 static void update_status(struct spice_connection *conn)
@@ -329,6 +339,41 @@ static void update_status(struct spice_connection *conn)
     }
 }
 
+static const char *spice_edit_properties[] = {
+    "CopyToGuest",
+    "PasteFromGuest",
+};
+
+static void update_edit_menu_window(SpiceWindow *win)
+{
+    int i;
+    GtkAction *toggle;
+
+    if (win == NULL) {
+        return;
+    }
+
+    /* Make "CopyToGuest" and "PasteFromGuest" insensitive if spice
+     * agent is not connected */
+    for (i = 0; i < G_N_ELEMENTS(spice_edit_properties); i++) {
+        toggle = gtk_action_group_get_action(win->ag, spice_edit_properties[i]);
+        if (toggle) {
+            gtk_action_set_sensitive(toggle, win->conn->agent_connected);
+        }
+    }
+}
+
+static void update_edit_menu(struct spice_connection *conn)
+{
+    int i;
+
+    for (i = 0; i < SPICE_N_ELEMENTS(conn->wins); i++) {
+        if (conn->wins[i]) {
+            update_edit_menu_window(conn->wins[i]);
+        }
+    }
+}
+
 static void menu_cb_connect(GtkAction *action, void *data)
 {
     struct spice_connection *conn;
@@ -380,13 +425,6 @@ static void menu_cb_fullscreen(GtkAction *action, void *data)
     window_set_fullscreen(win, !win->fullscreen);
 }
 
-static void menu_cb_ungrab(GtkAction *action, void *data)
-{
-    SpiceWindow *win = data;
-
-    spice_display_mouse_ungrab(SPICE_DISPLAY(win->spice));
-}
-
 #ifdef USE_SMARTCARD
 static void enable_smartcard_actions(SpiceWindow *win, VReader *reader,
                                      gboolean can_insert, gboolean can_remove)
@@ -760,13 +798,6 @@ static const GtkActionEntry entries[] = {
         .callback    = G_CALLBACK(menu_cb_fullscreen),
         .accelerator = "<shift>F11",
     },{
-
-        /* Input menu */
-        .name        = "UngrabMouse",
-        .label       = N_("_Ungrab mouse"),
-        .callback    = G_CALLBACK(menu_cb_ungrab),
-        .accelerator = "<shift>F12",
-    },{
 #ifdef USE_SMARTCARD
 	.name        = "InsertSmartcard",
 	.label       = N_("_Insert Smartcard"),
@@ -868,7 +899,6 @@ static char ui_xml[] =
 "      <menuitem action='Statusbar'/>\n"
 "    </menu>\n"
 "    <menu action='InputMenu'>\n"
-"      <menuitem action='UngrabMouse'/>\n"
 #ifdef USE_SMARTCARD
 "      <menuitem action='InsertSmartcard'/>\n"
 "      <menuitem action='RemoveSmartcard'/>\n"
@@ -1305,6 +1335,8 @@ static SpiceWindow *create_spice_window(spice_connection *conn, SpiceChannel *ch
                                       win, 0);
     }
 
+    update_edit_menu_window(win);
+
     toggle = gtk_action_group_get_action(win->ag, "Toolbar");
     state = gtk_widget_get_visible(win->toolbar);
     gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(toggle), state);
@@ -1462,6 +1494,7 @@ static void main_agent_update(SpiceChannel *channel, gpointer data)
     g_object_get(channel, "agent-connected", &conn->agent_connected, NULL);
     conn->agent_state = conn->agent_connected ? _("yes") : _("no");
     update_status(conn);
+    update_edit_menu(conn);
 }
 
 static void inputs_modifiers(SpiceChannel *channel, gpointer data)
@@ -1583,6 +1616,106 @@ static void display_monitors(SpiceChannel *display, GParamSpec *pspec,
     g_clear_pointer(&monitors, g_array_unref);
 }
 
+static void port_write_cb(GObject *source_object,
+                          GAsyncResult *res,
+                          gpointer user_data)
+{
+    SpicePortChannel *port = SPICE_PORT_CHANNEL(source_object);
+    GError *error = NULL;
+
+    spice_port_write_finish(port, res, &error);
+    if (error != NULL)
+        g_warning("%s", error->message);
+    g_clear_error(&error);
+}
+
+static void port_flushed_cb(GObject *source_object,
+                            GAsyncResult *res,
+                            gpointer user_data)
+{
+    SpiceChannel *channel = SPICE_CHANNEL(source_object);
+    GError *error = NULL;
+
+    spice_channel_flush_finish(channel, res, &error);
+    if (error != NULL)
+        g_warning("%s", error->message);
+    g_clear_error(&error);
+
+    spice_channel_disconnect(channel, SPICE_CHANNEL_CLOSED);
+}
+
+static gboolean input_cb(GIOChannel *gin, GIOCondition condition, gpointer data)
+{
+    char buf[4096];
+    gsize bytes_read;
+    GIOStatus status;
+
+    if (!(condition & G_IO_IN))
+        return FALSE;
+
+    status = g_io_channel_read_chars(gin, buf, sizeof(buf), &bytes_read, NULL);
+    if (status != G_IO_STATUS_NORMAL)
+        return FALSE;
+
+    if (stdin_port != NULL)
+        spice_port_write_async(stdin_port, buf, bytes_read, NULL, port_write_cb, NULL);
+
+    return TRUE;
+}
+
+static void port_opened(SpiceChannel *channel, GParamSpec *pspec,
+                        spice_connection *conn)
+{
+    SpicePortChannel *port = SPICE_PORT_CHANNEL(channel);
+    gchar *name = NULL;
+    gboolean opened = FALSE;
+
+    g_object_get(channel,
+                 "port-name", &name,
+                 "port-opened", &opened,
+                 NULL);
+
+    g_printerr("port %p %s: %s\n", channel, name, opened ? "opened" : "closed");
+
+    if (opened) {
+        /* only send a break event and disconnect */
+        if (g_strcmp0(name, "org.spice.spicy.break") == 0) {
+            spice_port_event(port, SPICE_PORT_EVENT_BREAK);
+        }
+
+        /* handle the first spicy port and connect it to stdin/out */
+        if (g_strcmp0(name, "org.spice.spicy") == 0 && stdin_port == NULL) {
+            stdin_port = port;
+            goto end;
+        }
+
+        if (port == stdin_port)
+            goto end;
+
+        spice_channel_flush_async(channel, NULL, port_flushed_cb, conn);
+    } else {
+        if (port == stdin_port)
+            stdin_port = NULL;
+    }
+
+end:
+    g_free(name);
+}
+
+static void port_data(SpicePortChannel *port,
+                      gpointer data, int size, spice_connection *conn)
+{
+    int r;
+
+    if (port != stdin_port)
+        return;
+
+    r = write(fileno(stdout), data, size);
+    if (r != size) {
+        g_warning("port write failed result %d/%d errno %d", r, size, errno);
+    }
+}
+
 static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
 {
     spice_connection *conn = data;
@@ -1630,6 +1763,14 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
     if (SPICE_IS_USBREDIR_CHANNEL(channel)) {
         update_auto_usbredir_sensitive(conn);
     }
+
+    if (SPICE_IS_PORT_CHANNEL(channel)) {
+        g_signal_connect(channel, "notify::port-opened",
+                         G_CALLBACK(port_opened), conn);
+        g_signal_connect(channel, "port-data",
+                         G_CALLBACK(port_data), conn);
+        spice_channel_connect(channel);
+    }
 }
 
 static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer data)
@@ -1658,6 +1799,11 @@ static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer dat
         update_auto_usbredir_sensitive(conn);
     }
 
+    if (SPICE_IS_PORT_CHANNEL(channel)) {
+        if (SPICE_PORT_CHANNEL(channel) == stdin_port)
+            stdin_port = NULL;
+    }
+
     conn->channels--;
     if (conn->channels > 0) {
         return;
@@ -1810,6 +1956,40 @@ static void usb_connect_failed(GObject               *object,
     gtk_widget_destroy(dialog);
 }
 
+static void setup_terminal(gboolean reset)
+{
+    int stdinfd = fileno(stdin);
+
+    if (!isatty(stdinfd))
+        return;
+
+#ifdef HAVE_TERMIOS_H
+    static struct termios saved_tios;
+    struct termios tios;
+
+    if (reset)
+        tios = saved_tios;
+    else {
+        tcgetattr(stdinfd, &tios);
+        saved_tios = tios;
+        tios.c_lflag &= ~(ICANON | ECHO);
+    }
+
+    tcsetattr(stdinfd, TCSANOW, &tios);
+#endif
+}
+
+static void watch_stdin(void)
+{
+    int stdinfd = fileno(stdin);
+    GIOChannel *gin;
+
+    setup_terminal(false);
+    gin = g_io_channel_unix_new(stdinfd);
+    g_io_channel_set_flags(gin, G_IO_FLAG_NONBLOCK, NULL);
+    g_io_add_watch(gin, G_IO_IN|G_IO_ERR|G_IO_HUP|G_IO_NVAL, input_cb, NULL);
+}
+
 int main(int argc, char *argv[])
 {
     GError *error = NULL;
@@ -1900,6 +2080,8 @@ int main(int argc, char *argv[])
     g_free(port);
     g_free(tls_port);
 
+    watch_stdin();
+
     connection_connect(conn);
     if (connections > 0)
         g_main_loop_run(mainloop);
@@ -1921,5 +2103,6 @@ int main(int argc, char *argv[])
 
     g_free(spicy_title);
 
+    setup_terminal(true);
     return 0;
 }
